using System;
using System.Text.RegularExpressions;

namespace WebToolkit
{
    /// <summary>
    /// Class that encapsulates BBcode related functionality
    /// </summary>
    public static class BBcode
    {
        /// <summary>
        /// Transforms BBcode formated text into XHTML
        /// </summary>
        /// <param name="text">BBcode formatted text</param>
        /// <returns>XHTML-encoded text</returns>
        private static string BBdecodeInternal(string input)
        {
			string original = input;

            // BB-code stuff

			// [b]stuff[/b]
			string rxStylesB = @"\[b\]([^\]]*)\[\/b\]";
            input = Regex.Replace(input, rxStylesB, "<b>$1</b>");

			// [i]stuff[/i]
            string rxStylesI = @"\[i\]([^\[]*)\[\/i\]";
            input = Regex.Replace(input, rxStylesI, "<i>$1</i>");

			// [u]stuff[/i]
            string rxStylesU = @"\[u\]([^\[]*)\[\/u\]";
            input = Regex.Replace(input, rxStylesU, "<u>$1</u>");

			// [code]stuff[/stuff]
            string rxCode = @"\[code\]([^\[]*)\[\/code\]";
            input = Regex.Replace(input, rxCode, "</p><pre>$1</pre><p>");

			// [img src]alt[/img]
            string rxImages = @"\[img[ |=]([^\]]*)\]([^\[]*)\[\/img\]";
            input = Regex.Replace(input, rxImages, "<img src=\"$1\" alt=\"$2\"/>");

			// [img]src[/img]
			string rxImageNoAlt = @"\[img\]([^\[]*)\[\/img\]";
			input = Regex.Replace(input, rxImageNoAlt, "<img src=\"$1\" alt=\"\"/>");

			// [img src]
			string rxImageDirect = @"\[img[ |=]([^\]]*)\]";
			input = Regex.Replace(input, rxImageDirect, "<img src=\"$1\" alt=\"\"/>");

			// [affil link]text[/affil]
            string rxAffiliate = @"\[affil (http:\/\/)?([^\]]*)\]([^\[]*)\[\/affil\]";
            input = Regex.Replace(input, rxAffiliate, "<a href=\"http://global.yesasia.com/assocred.asp?6LTLNB7Q+http://$2\">$3</a>");

			// [url]link[/url]
            string rxLink1 = @"\[url\](http:\/\/)?([^\[]*)\[\/url\]";
            input = Regex.Replace(input, rxLink1, "<a href=\"/go.ashx?http://$2\">$2</a>");

			// [url link]text[/url]
            string rxLink2 = @"\[url[ |=](http:\/\/)?([^\]]*)\]([^\[]*)\[\/url\]";
            input = Regex.Replace(input, rxLink2, "<a href=\"/go.ashx?http://$2\">$3</a>");

			// [quote]stuff[/quote]
			string rxQuoteAnon = @"\[quote\]([^\]]*)\[\/quote\]";
			input = Regex.Replace(input, rxQuoteAnon, "<blockquote>$1</blockquote>");

            // handle lists
            string rxListMain = @"\s*\[list\]([\s]*)([^\[]*)\[/list\](\s*)";
            input = Regex.Replace(input, rxListMain, "<ul>$2</ul>");
            string rxListItem = @"(--([^\n]+)\s*)";
            input = Regex.Replace(input, rxListItem, "<li>$2</li>");

			string rxMore = @"(\[more\](\n?))";
			input = Regex.Replace(input, rxMore, "");

			if (input == original)
			{
				return input;
			}
			else
			{
				return BBdecodeInternal(input);
			}
		}

		/// <summary>
		/// Transforms BBcode formated text into XHTML
		/// </summary>
		/// <param name="text">BBcode formatted text</param>
		/// <returns>XHTML-encoded text</returns>
		public static string BBdecode(string input)
		{
			// fix HTML entitites and disable arbitrary HTML code
			//input = input.Replace("&", "&amp;");
			//input = input.Replace("<", "&lt;");
			//input = input.Replace(">", "&gt;");

			input = System.Web.HttpContext.Current.Server.HtmlEncode(input);

			input = BBdecodeInternal(input);

			// to be able to write [something] statements without actually using
            // the bbCode
            input = input.Replace(@"[-", "[");

            // make newlines into breaks
            input = Regex.Replace(input, "(\n(\r{0,1}))", "<br />");

			// smileys!
			input = Smiley.Apply(input);

            return input;
        }

        /// <summary>
        /// Takes BBcode formatted text and shortens it based on the
        /// number of visible characters after BBcode is stripped/rendered.
        /// Output will be paragraph nearest to specified length.
        /// </summary>
        /// <param name="original">BBcode foramtted text</param>
        /// <param name="length">Desired outputlength</param>
        /// <returns>BBcode formatted, shortened text</returns>
        public static string Shorten(string original, int length)
        {
            string input = original;
            string text = input;

            int introSplitPos = text.IndexOf("[more]");
            if (introSplitPos != -1)
            {
                text = text.Substring(0, introSplitPos-1);
                text += "\n[i]More...[/i]";
            }
            else if (textLength(text) > length)
            {
                text = "";
                int pos = 0;
                int prepos = 0;
                int realLength = 0;

                while (realLength <= length && pos != -1)
                {
                    prepos = pos;
                    pos = input.IndexOf('\n', prepos + 1);

                    if (pos != -1)
                    {
                        string segment = input.Substring(prepos, (pos - prepos));
                        int segmentLength = textLength(segment);

                        if ((realLength + segmentLength) <= length)
                        {
                            text += segment;
                            realLength += textLength(segment);
                        }
                        else
                        {
                            int diff1 = length - realLength;
                            int diff2 = (realLength + segmentLength) - length;

                            if (diff1 > diff2)
                            {
                                text += segment;
                                realLength += textLength(segment);
                            }
                            else
                            {
                                // add to length-counter but do not add content
                                realLength += textLength(segment);
                            }
                        }
                    }
                }

                if (text.Length != input.Length)
                {
                    text += "\n[i]More...[/i]";
                }
            }

            return text;
        }

        private static int textLength(string text)
        {
            string rxMatch = @"\[([^\]]*)\]";
            string rxReplace = "";
            Regex rx = new Regex(rxMatch);
            string filtered = rx.Replace(text, rxReplace);

            return filtered.Length;
        }
    }
}