﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace Sayumin.Common
{
    /// <summary>
    /// Class which implements rudimentary, overridable scan-mechanisms
    /// </summary>
    public class BaseScanner : ConfigReader, IRadioScanner
    {
        protected delegate bool OnMatchDelegate(FileEntry Entry, string InputValue);


        /// <summary>
        /// Creates a new BaseScanner object.
        /// </summary>
        public BaseScanner() : base()
        {
        }

        /// <summary>
        /// Initialises the settings based on DB-configuration
        /// </summary>
        /// <param name="Settings">Settings should be a key/pair-values seperated by semicolons with equality for assignemnt.</param>
        public override void InitSettings(string Settings)
        {
            const string defaultDateMatch = @"(?<Y>\d{2})(?<M>\d{2})(?<D>\d{2})";
            _settings.Add("dateformat", defaultDateMatch);
            _settings.Add("extension", "mp3");

            base.InitSettings(Settings);
        }

        #region IRadioScanner Members

        public FileEntry FindMatch(FileEntry[] Entries, DateTime EpisodeDate)
        {
            if (Entries == null)
            {
                return null;
            }

            FileEntry match = null;

            foreach (FileEntry entry in Entries)
            {
                bool isMatch = MatchEntry(entry, EpisodeDate);
                if (isMatch)
                {
                    match = entry;
                    break;
                }
            }

            return match;
        }

        protected virtual bool MatchEntry(FileEntry Entry, DateTime EpisodeDate)
        {
            bool isMatch = true;

            isMatch &= Match(Entry, "Desc", (e, x) => MatchDescription(e, x));
            isMatch &= Match(Entry, "MinSize", (e, x) => MatchSize(e, x));
            isMatch &= Match(Entry, "Extension", (e, x) => MatchExtension(e, x));
            isMatch &= MatchDate(Entry, EpisodeDate);
            isMatch &= MatchAdditional(Entry);

            return isMatch;
        }

        /// <summary>
        /// Checks if settings are present for the provided key.
        /// Will always return match if no key is present in config data.
        /// </summary>
        /// <param name="Entry">The FileEntry object to search for match.</param>
        /// <param name="Key"></param>
        /// <param name="MatchMethod"></param>
        /// <returns></returns>
        protected bool Match(FileEntry Entry, string Key, OnMatchDelegate MatchMethod)
        {
            if (false == _settings.Keys.Contains(Key.ToLower()))
            {
                return true;
            }

            string value = _settings[Key.ToLower()];
            return MatchMethod(Entry, value);
        }

        protected virtual bool MatchDescription(FileEntry Entry, string Description)
        {
            return Entry.Description.Contains(Description);
        }

        private bool MatchSize(FileEntry Entry, string FileSize)
        {
            long fileSize = Convert.ToInt64(FileSize);
            return MatchSize(Entry, fileSize);
        }

        protected virtual bool MatchSize(FileEntry Entry, long FileSize)
        {
            return Entry.FileSize >= FileSize;
        }

        protected DateTime GetDate(string Content)
        {
            DateTime detectedDate = DateTime.MinValue;
            Match m = Regex.Match(Content, _settings["dateformat"]);

            if (m.Success)
            {
                string yy = m.Groups["Y"].Value;
                string mm = m.Groups["M"].Value;
                string dd = m.Groups["D"].Value;

                if (yy.Length == 2)
                {
                    yy = "20" + yy;
                }

                string dateString = string.Format("{0}-{1}-{2}", yy, mm, dd);
                detectedDate = DateTime.Parse(dateString);
            }

            return detectedDate;
        }

        protected virtual bool MatchDate(FileEntry Entry, DateTime EpisodeDate)
        {
            return EpisodeDate.Date == GetDate(Entry.Description);
                //&& EpisodeDate >= Entry.UploadDate.Date;
        }

        protected virtual bool MatchExtension(FileEntry Entry, string Extension)
        {
            return Entry.Extension.ToLower() == Extension.ToLower();
        }

        protected virtual bool MatchAdditional(FileEntry Entry)
        {
            return true;
        }

        #endregion
    }
}
