﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace Sayumin.Common
{
    public class BaseSource : ConfigReader, IFileSource
    {
        /*
         * network variables
         */
        protected ScreenScraper _scraper;
        protected DateTime _lastUpdated;

        /*
         * used for pre-processed input-data
         */
        protected string[,] _rowData;
        protected string[] _headers;

        /*
         * used for column-detection and setup
         */
        protected List<int> _unusedColumns;
        protected int _mediacol;
        protected int _desccol;
        protected int _sizecol;
        protected int _datecol;
        protected int _linkOffset;

        /*
         * constants for header detection
         */
        protected string[] H_MEDIACOL = new string[] { "name", "dl"};
        protected string[] H_DESCCOL = new string[] { "comment", "説明", "description" };
        protected string[] H_SIZECOL = new string[] { "size", "サイズ" };
        protected string[] H_DATECOL = new string[] { "date" };

        /// <summary>
        /// Reports when the source was last checked.
        /// </summary>
        public DateTime LastUpdated
        {
            get
            {
                return _lastUpdated;
            }
            set
            {
                _lastUpdated = value;
            }
        }

        /*
         * constructor
         */
        public BaseSource()
            : base()
        {
        }

        /*
         * setup code
         */

        /// <summary>
        /// Sets default settings and reads any additional settings specified.
        /// </summary>
        public override void InitSettings(string Settings)
        {
            // set defaults before loading user-specified.

            // japanese time
            // format: 2009/01/10(Sat) 17:18
            _settings["dateformat"] = "(?<Y>\\d{4})\\/(?<M>\\d{2})\\/(?<D>\\d{2})\\([A-Za-z]+\\)( ?)(?<h>\\d{2}):(?<m>\\d{2})";

            // size-parameter ignorered until there is a plan here
            // assume one of the two following formats:
            // 93.15MB / 17KB
            // 93,912,631

            _settings["encoding"] = "EUC-JP";
            _settings["referer"] = string.Empty;
            _settings["startdelimiter"] = "<th";
            _settings["stopdelimiter"] = "</table>";
            _settings["tagdelimiter"] = "\"";
            _settings["linkoffset"] = "0"; // used for specifying offset in path for links to mediafiles 

            // no default for "url", "linkprefix"

            // reset all other settings:
            _settings["url"] = null;
            _settings["referrer"] = null;
            _settings["linkprefix"] = null;

            base.InitSettings(Settings);
        }

        /// <summary>
        /// Initializes and configures the screen-scraper and sets required properties.
        /// </summary>
        protected virtual void InitScraper()
        {
            // get settings
            string url = _settings["url"];
            string referer = _settings["referrer"];
            string startDelimiter = _settings["startdelimiter"];
            string stopDelimiter = _settings["stopdelimiter"];
            string tagDelimiter = _settings["tagdelimiter"];

            _scraper = new ScreenScraper(url, "http://www.sayumin.org", referer);
            _scraper.StartDelimiter = startDelimiter;
            _scraper.StopDelimiter = stopDelimiter;
            _scraper.TagDelimiter = tagDelimiter;

            Encoding encoding = Encoding.GetEncoding(_settings["encoding"]);
            _scraper.Encoding = encoding;

            /* 
             * not set by default: (integers for column-numbers with specific data)
             * mediacol
             * sizecol
             * datecol
             * desccol
             * - will be autodetected based on headers
             */
        }

        /*
         * workflow
         */

        /// <summary>
        /// Gets the actual file-entries from the source
        /// </summary>
        /// <returns>An array of all files detected at source.</returns>
        public virtual FileEntry[] GetEntries()
        {
            _lastUpdated = DateTime.MinValue;
            FileEntry[] entries = null;

            InitScraper();

            _linkOffset = int.Parse(_settings["linkoffset"]);

            FetchData();
            GetColumnConfig();
            entries = GenerateFileEntryList();

            // only claim updated if successfull
            _lastUpdated = DateTime.Now;
            return entries;
        }

        /// <summary>
        /// Fetches data and constructs the needed header and
        /// table structures.
        /// </summary>
        protected virtual void FetchData()
        {
            _headers = null;
            _rowData = null;

            _scraper.Fetch();

            if (!_scraper.IsLoaded)
            {
                return;
            }

            this.ReformatData();

            // analyze headers

            List<string> headers = new List<string>();
            while (false == _scraper.IsComplete)
            {
                _scraper.Skip(">", true);
                string header = _scraper.Cut("</th>");

                if (false == string.IsNullOrEmpty(header))
                {
                    headers.Add(header.ToLower());
                }

                _scraper.Skip("<th");
            }
            _headers = headers.ToArray();

            _scraper.Rewind();
            _scraper.Skip("<tr");
            
            _rowData = _scraper.GetTableData();
        }

        /// <summary>
        /// Sets up a internal table-definition for columns and related data
        /// based on configuration-data and _headers, in that order.
        /// </summary>
        protected virtual void GetColumnConfig()
        {
            _unusedColumns = new List<int>();

            if (_rowData == null)
            {
                return;
            }

            //int numRows = _rowData.GetLength(0);
            int numColumns = _headers.Length;

            //
            // setup column-mapping
            //

            // column table
            for (int i = 0; i < numColumns; i++)
            {
                _unusedColumns.Add(i);
            }

            // get explicitly configured columns

            _mediacol = GetColumn("media");
            _desccol = GetColumn("desc");
            _sizecol = GetColumn("size");
            _datecol = GetColumn("date");

            int[] configColumns = new int[] {
                _mediacol, _desccol, _sizecol, _datecol
            };

            foreach (int column in configColumns)
            {
                if (column != -1)
                {
                    _unusedColumns.Remove(column);
                }
            }

            // autodetermine any remaining columns based on headers

            _mediacol = CheckColumn(_mediacol, H_MEDIACOL);
            _desccol = CheckColumn(_desccol, H_DESCCOL);
            _datecol = CheckColumn(_datecol, H_DATECOL);
            _sizecol = CheckColumn(_sizecol, H_SIZECOL);

            // everything remaining in _unusedColumns are "additionals"
        }

        /// <summary>
        /// Generates a list of file-entry objects based on the existing pre-processed site-data.
        /// </summary>
        private FileEntry[] GenerateFileEntryList()
        {
            if (_rowData == null)
            {
                return null;
            }

            List<FileEntry> entries = new List<FileEntry>();
            for (int i = 0; i < _rowData.GetLength(0); i++)
            {
                string mediaRaw = GetTableCell(i, _mediacol);
                string descRaw = GetTableCell(i, _desccol);
                string sizeRaw = GetTableCell(i, _sizecol);
                string dateRaw = GetTableCell(i, _datecol);

                descRaw = EliminiateElements(descRaw);
                sizeRaw = EliminiateElements(sizeRaw);
                dateRaw = EliminiateElements(dateRaw);

                string url = GetMediaLink(mediaRaw);
                string extension = GetMediaExtension(mediaRaw);
                string desc = GetDescription(descRaw);
                long size = GetSize(sizeRaw);
                DateTime uploadData = GetDate(dateRaw);
                Dictionary<string, string> additionals = GetAdditionals(i);

                FileEntry entry = new FileEntry();
                entry.Url = url;
                entry.Extension = extension;
                entry.Description = desc;
                entry.FileSize = size;
                entry.UploadDate = uploadData;
                entry.Columns = additionals;

                entries.Add(entry);
            }

            FileEntry[] fileEntries = entries.ToArray();
            return fileEntries;
        }

        /*
         * overideable workflow components
         */

        /// <summary>
        /// Overridable method which allows reformating of raw HTML-data
        /// prior to any table-processing.
        /// </summary>
        protected virtual void ReformatData()
        {
        }

        /*
         * overridable parsing-methods for specific, derived sources with mixed content
         * to allow custom processing logic.
         */

        /// <summary>
        /// Return the URL for the media-entry.
        /// </summary>
        protected virtual string GetMediaLink(string RawData)
        {
            if (RawData == null)
            {
                return null;
            }

            if (RawData.StartsWith("["))
            {
                RawData = RawData.Substring(1);
            }

            string url = _scraper.GetLinkedURL(RawData);

            if (_settings["linkprefix"] != null)
            {
                url = _settings["linkprefix"] + url.Substring(_linkOffset);
            }
            else
            {
                url = GetAbsoluteUrl(url);
            }
            return url;
        }

        /// <summary>
        /// Return the visually reported name for the media-entry.
        /// </summary>
        protected virtual string GetMediaExtension(string RawData)
        {
            if (RawData == null)
            {
                return null;
            }

            if (RawData.StartsWith("["))
            {
                RawData = RawData.Substring(1);
            }

            string name = _scraper.GetLinkedText(RawData);

            for (int i = name.Length; i != 0; i--)
            {
                if (name[i-1] == '.')
                {
                    name = name.Substring(i);
                    break;
                }
            }
            return name;
        }

        /// <summary>
        /// Parses the size-parameter and returns it in numeric form
        /// </summary>
        protected virtual long GetSize(string RawData)
        {
            if (RawData == null)
            {
                return default(long);
            }

            // two know formats:
            // 1,234,456,345
            // 22.45MB / 17KB

            RawData = RawData.ToLower().Trim();

            long size = default(long);
            if (RawData.Contains("b"))
            {
                int scalePos = RawData.IndexOf('b');

                string numData = RawData.Substring(0, scalePos - 1);

                string seperator = System.Threading.Thread.CurrentThread.CurrentCulture.NumberFormat.NumberDecimalSeparator;
                numData = numData.Replace(",", seperator);
                numData = numData.Replace(".", seperator);
                decimal unscaledSize = decimal.Parse(numData);

                unscaledSize = RoundDown(unscaledSize, numData);

                string scaleUnit = RawData.Substring(scalePos - 1, 1);

                switch (scaleUnit)
                {
                    case "k":
                        unscaledSize *= 1024;
                        break; ;

                    case "m":
                        unscaledSize *= 1024 * 1024;
                        break; ;

                    case "g":
                        unscaledSize *= 1024 * 1024 * 1024;
                        break; ;
                }

                size = Convert.ToInt64(unscaledSize);
            }
            else
            {
                // plain size in bytes
                RawData = RawData.Replace(",", "");
                RawData = RawData.Replace(".", "");

                size = long.Parse(RawData);
            }

            return size;
        }

        /// <summary>
        /// Returnd the reported upload-date.
        /// </summary>
        protected virtual DateTime GetDate(string RawData)
        {
            DateTime uploadDate = default(DateTime);

            if (RawData != null)
            {

                Match m = Regex.Match(RawData, _settings["dateformat"]);

                if (m.Success)
                {

                    string yy = m.Groups["Y"].Value;
                    string MM = m.Groups["M"].Value;
                    string dd = m.Groups["D"].Value;

                    string hh = string.Empty;
                    string mm = string.Empty;

                    try
                    {
                        hh = m.Groups["h"].Value;
                        mm = m.Groups["m"].Value;
                    }
                    catch { }

                    string dateString;
                    if (false == string.IsNullOrEmpty(mm))
                    {
                        dateString = string.Format("{0}-{1}-{2} {3}:{4}", yy, MM, dd, hh, mm);
                    }
                    else
                    {
                        dateString = string.Format("{0}-{1}-{2}", yy, MM, dd);
                    }

                    // japan = utc +9
                    const int hoursOffUtc = 9;

                    uploadDate = DateTime.Parse(dateString);
                    uploadDate = ConvertFromRandomToLocalTime(uploadDate, hoursOffUtc);
                }
            }
            return uploadDate;
        }

        /// <summary>
        /// Returns the reported description.
        /// </summary>
        protected virtual string GetDescription(string RawData)
        {
            return RawData;
        }

        /// <summary>
        /// Returns the additional columns containing additional data.
        /// </summary>
        /// <returns></returns>
        protected virtual Dictionary<string,string> GetAdditionals(int row)
        {
            Dictionary<string, string> additionals = new Dictionary<string, string>();

            foreach (int column in _unusedColumns)
            {
                additionals.Add(_headers[column], _rowData[row, column]);
            }

            return additionals;
        }

        /*
         * various helper methods
         */

        /// <summary>
        /// Converts time from a random timezone to whatever local time is.
        /// </summary>
        private DateTime ConvertFromRandomToLocalTime(DateTime Source, int HoursOffUtc)
        {
            DateTime kindAdjusted = DateTime.SpecifyKind(Source, DateTimeKind.Utc);
            kindAdjusted = kindAdjusted.AddHours(-HoursOffUtc);
            DateTime converted = TimeZone.CurrentTimeZone.ToLocalTime(kindAdjusted);
            return converted;
        }

        /// <summary>
        /// Check in Downloader-component will check if file is size X or larger
        /// and consider the download failed if the size is below this.
        /// This method rounds down decimal-numbers which might have rounded up in order to be safe about the download size limit.
        /// </summary>
        private decimal RoundDown(decimal currentValue, string stringRepresentation)
        {
            int pos = stringRepresentation.IndexOf(System.Threading.Thread.CurrentThread.CurrentCulture.NumberFormat.NumberDecimalSeparator);

            if (pos == -1)
            {
                currentValue -= (decimal)0.5;
                return currentValue;
            }

            int numDecimals = stringRepresentation.Length - pos - 1;

            decimal roundDownValue = (decimal)(0.5 / (Math.Pow(10, numDecimals)));
            currentValue -= roundDownValue;
            return currentValue;
        }

        /// <summary>
        /// Combines the portions of data from the url-setting and the retrieved link to make an absolute URL
        /// </summary>
        private string GetAbsoluteUrl(string FileUrl)
        {
            string url = _settings["url"];

            string urlRoot = Regex.Replace(url, @"(http:\/\/[a-z0-9\.\-]+)\/(.+)", "$1");
            string urlFolder = Regex.Replace(url, @"(http:\/\/.*\/)([^\/]+)", "$1");

            string absoluteUrl;

            if (FileUrl.StartsWith("/"))
            {
                absoluteUrl = urlRoot + FileUrl;
            }
            else if (FileUrl.StartsWith("./"))
            {
                absoluteUrl = urlFolder + FileUrl.Substring(2);
            }
            else if (FileUrl.StartsWith("../"))
            {
                string urlFolderParent = Regex.Replace(urlFolder, @"(http:\/\/.*\/)([^\/]+)/", "$1");
                absoluteUrl = urlFolderParent + FileUrl.Substring(3);
            }
            else
            {
                absoluteUrl = urlFolder + FileUrl;
            }

            return absoluteUrl;
        }

        /// <summary>
        /// Gets the column-number based on config data.
        /// </summary>
        /// <param name="key">column-key without the col-suffix</param>
        /// <returns>-1 if not found</returns>
        private int GetColumn(string key)
        {
            key += "col";
            if (false == _settings.Keys.Contains(key))
            {
                return -1;
            }

            return int.Parse(_settings["key"]);
        }

        /// <summary>
        /// Checks if the column is already defined and if not searches
        /// the available headers for a valid match.
        /// If a valid match is found, the match is remove from _unusedHeaders.
        /// </summary>
        /// <param name="columnNum">The columnindex to search</param>
        /// <param name="validHeaders">a string-array of valid headers</param>
        /// <returns>Returns columnNum if not -1, -1 if valid header is not found
        /// or the index of the valid match.</returns>
        private int CheckColumn(int columnNum, string[] validHeaders)
        {
            if (columnNum != -1)
            {
                return columnNum;
            }

            foreach (string validHeader in validHeaders)
            {
                for (int i = 0; i < _headers.Length; i++)
                {
                    if (_headers[i] == validHeader)
                    {
                        _unusedColumns.Remove(i);
                        return i;
                    }
                }
            }

            return -1;
        }

        /// <summary>
        /// Does a simple check on column-number and returns null if invalid,
        /// otherwise data is retrieved from the table.
        /// </summary>
        private string GetTableCell(int row, int columnNum)
        {
            if (columnNum == -1)
            {
                return null;
            }

            return _rowData[row, columnNum];
        }

        /// <summary>
        /// Removes surrounding HTML-elements and returns the contained text
        /// </summary>
        private string EliminiateElements(string RawData)
        {
            string reformatted = string.Empty;
            int pos = RawData.IndexOf("</");

            if (pos != -1)
            {
                RawData = RawData.Substring(0, pos);
            }

            pos = RawData.LastIndexOf(">");

            if (pos != -1)
            {
                RawData = RawData.Substring(pos + 1);
            }
            return RawData;
        }
    }
}