﻿using System;
using System.Text;
using System.Net;
using System.Collections.Generic;

namespace Sayumin.Common
{
    /// <summary>
    /// A class containing HTML screen-scraping related functionality
    /// </summary>
    public class ScreenScraper
    {
        // configurable data
        private string _url;
        private string _startdelimiter;
        private string _stopdelimiter;
        private string _tagDelimiter;

        private bool _loaded;
        private bool _complete;
        
        private int _currentPosition;
        private List<int> _oldPositions;

        private int _position
        {
            get
            {
                return _currentPosition;
            }
            set
            {
                _oldPositions.Insert(0, _currentPosition);
                _currentPosition = value;
            }
        }

        private string _data;

        private string _userAgent = "Jostein Kjønigsen's ScreenScraper library";

        /*
         * public properties
         */

        public string Url
        {
            get { return _url; }
            set { _url = value; }
        }

        private Encoding _encoding;

        /// <summary>
        /// Alternate encoding used instead of UTF8 to retrieve data
        /// </summary>
        public Encoding Encoding
        {
            get { return _encoding; }
            set { _encoding = value; }
        }

        /// <summary>
        /// Match-string which determines where to start processing the document
        /// </summary>
        public string StartDelimiter
        {
            set { _startdelimiter = value; }
            get { return _startdelimiter; }
        }

        /// <summary>
        /// Match-string which determines where to stop processing the document
        /// </summary>
        public string StopDelimiter
        {
            set { _stopdelimiter = value; }
            get { return _stopdelimiter; }
        }

        /// <summary>
        /// Determines if data has been loaded succesfully
        /// </summary>
        public bool IsLoaded
        { get { return _loaded; } }

        /// <summary>
        /// Indicates if document is completely parsed
        /// </summary>
        public bool IsComplete
        { get { return _complete; } }

        /// <summary>
        /// The internally stored fetched data. Modifying this will not
        /// reallign the positioning, so this should be done before any
        /// other processing.
        /// </summary>
        public string Data
        {
            get { return _data; }
            set { _data = value; }
        }

        public string Referer { get; set; }

        public string TagDelimiter
        {
            get { return _tagDelimiter; }
            set { _tagDelimiter = value; }
        }

        /// <summary>
        /// Creates a new instance of the Screenscraper object
        /// </summary>
        /// <param name="url">URL to be scraped</param>
        public ScreenScraper(string url)
        {
            _url = url;
            Referer = "";
            _oldPositions = new List<int>();
        }

        /// <summary>
        /// Creates a new instance of the Screenscraper object
        /// </summary>
        /// <param name="url">URL to be scraped</param>
        /// <param name="reference">Reference URL to insert in user-agent</param>
        public ScreenScraper(string url, string reference)
            : this(url)
        {
            _userAgent += " (" + reference + ")";
        }

        /// <summary>
        /// Creates a new instance of the Screenscraper object
        /// </summary>
        /// <param name="url">URL to be scraped</param>
        /// <param name="reference">Reference URL to insert in user-agent</param>
        public ScreenScraper(string url, string reference, string referer)
            : this(url, reference)
        {
            Referer = referer;
        }

        /// <summary>
        /// Loads data into memory for processing
        /// </summary>
        public void Fetch()
        {
            _loaded = false;

            try
            {
                WebClient wc = new WebClient();
                wc.Headers.Add("User-Agent: " + _userAgent);
                wc.Encoding = System.Text.Encoding.UTF8;

                if (_encoding != null)
                {
                    wc.Encoding = _encoding;
                }

                if (Referer != "")
                {
                    wc.Headers.Add("Referer: " + Referer);
                }

                string tempdata = wc.DownloadString(_url);

                int start = 0;
                if (_startdelimiter != null)
                {
                    start = tempdata.IndexOf(_startdelimiter); // +_startdelimiter.Length;
                }

                int stop = tempdata.Length;
                if (_stopdelimiter != null)
                {
                    stop = tempdata.IndexOf(_stopdelimiter, start);
                }

                _data = tempdata.Substring(start, (stop - start));
                tempdata = null;

                _loaded = true;

                _position = 0;
            }
            catch (Exception) { }
        }

        /*
         * Stream processing
         * Forward-only
         */

        /// <summary>
        /// Skips position forward to the first match of the specified string
        /// </summary>
        /// <param name="match">Text to match</param>
        public void Skip(string match)
        {
            Skip(match, false);
        }

        /// <summary>
        /// Skips position forward to the first match of the specified string
        /// </summary>
        /// <param name="match">Text to match</param>
        /// <param name="SkipPast">Skip past the matched and set position to this point</param>
        public void Skip(string match, bool SkipPast)
        {
            if (_complete)
                return;

            _position = _data.IndexOf(match, _position + 1);
            if (_position == -1)
            {
                _complete = true;
                return;
            }

            if (SkipPast)
            {
                _position += match.Length;

                if (_position > _data.Length)
                {
                    _complete = true;
                }
            }
        }

        /// <summary>
        /// Skips the cursor back to the former location. Repeatable.
        /// <returns>true if successfull. false if there are no more to rewind.</returns>
        /// </summary>
        public bool Rewind()
        {
            if (_oldPositions.Count == 0)
            {
                return false;
            }

            _currentPosition = _oldPositions[0];
            _oldPositions.RemoveAt(0);

            return true;
        }

        /// <summary>
        /// Cuts out the text from the current position up until the first match and skips position forward
        /// </summary>
        /// <param name="match">Match for end of region to cut</param>
        /// <returns></returns>
        public string Cut(string match)
        {
            string cutData = null;

            int startPosition = _position;
            Skip(match);

            if (!_complete)
            {
                cutData = _data.Substring(startPosition, (_position - startPosition));
            }

            return cutData;
        }

        /// <summary>
        /// Gets the inner content of the first matching HTML-tag with any attributes
        /// </summary>
        /// <param name="tag">HTML-tag to find</param>
        /// <returns>Inner content of the specified tag</returns>
        public string GetTag(string tag)
        {
            return GetTag(tag, null);
        }

        /// <summary>
        /// Gets the inner content of the first matching HTML-tag with the specified attributes
        /// </summary>
        /// <param name="tag">HTML-tag to find</param>
        /// <param name="attributes">attributes with the HTML-tag</param>
        /// <returns>Inner content of the specified tag</returns>
        public string GetTag(string tag, string attributes)
        {
            if (_complete)
                return null;

            try
            {
                string startMatch = "<" + tag;
                if (attributes != null)
                {
                    startMatch += " " + attributes;
                }

                int start = _data.IndexOf(startMatch, _position);
                start = _data.IndexOf(">", start) + 1;

                int stop = _data.IndexOf("</" + tag + ">", start + 1);

                _position = stop;

                return _data.Substring(start, (stop - start));
            }
            catch (ArgumentOutOfRangeException)
            {
                _complete = true;
                return null;
            }
        }

        /// <summary>
        /// Gets the full HTML elemenet from the first match
        /// </summary>
        /// <param name="tag">HTML-tag to find</param>
        /// <returns>Inner content of the specified tag</returns>
        public string GetElement(string tag)
        {
            return GetElement(tag, null);
        }


        /// <summary>
        /// Gets the full HTML elemenet from the first match with the specified attributes
        /// </summary>
        /// <param name="tag">HTML-tag to find</param>
        /// <param name="attributes">attributes with the HTML-tag</param>
        /// <returns>Inner content of the specified tag</returns>
        public string GetElement(string tag, string attributes)
        {
            if (_complete)
                return null;

            try
            {
                string startMatch = "<" + tag;
                if (attributes != null)
                {
                    startMatch += " " + attributes;
                }

                int start = _data.IndexOf(startMatch, _position);
                int startInnerHTML = _data.IndexOf(">", start) + 1;

                string stopMatch = string.Format("</{0}>", tag);
                int stop = _data.IndexOf(stopMatch, startInnerHTML + 1) + stopMatch.Length;

                _position = stop - 1;

                return _data.Substring(start, (stop - start));
            }
            catch (ArgumentOutOfRangeException)
            {
                _complete = true;
                return null;
            }
        }

        /// <summary>
        /// Gets the contents of the first matching H2 HTML-element
        /// </summary>
        /// <returns>The inner content of the H2 HTML-element</returns>
        public string GetHeader()
        {
            return GetTag("h2");
        }

        /// <summary>
        /// Gets the contents of the first DIV-element
        /// </summary>
        /// <returns>The inner content of the DIV HTML-element</returns>
        public string GetDiv()
        {
            return GetTag("div");
        }

        /// <summary>
        /// Gets the contents of the first matching DIV HTML-element
        /// </summary>
        /// <returns>The inner content of the DIV HTML-element</returns>
        public string GetDiv(string cssClass)
        {
            return GetTag("div", "class=\"" + cssClass + "\"");
        }

        /// <summary>
        /// Gets the content of the first matching table-cell
        /// </summary>
        /// <returns>Inner-content of the table-cell element</returns>
        public string GetCell()
        {
            return GetTag("td");
        }

        /*
         * Stream-indendent methods
         */

        /// <summary>
        /// Returns the linked text in a HTML anchor/link element
        /// </summary>
        /// <param name="text">Full HTML Anchor/link element</param>
        /// <returns>The linked text</returns>
        public string GetLinkedText(string text)
        {
            string matchtext = text.ToLower();
            int start = matchtext.IndexOf("<a");
            start = matchtext.IndexOf(">", start) + 1;
            int stop = matchtext.IndexOf("</a>", start);

            return text.Substring(start, (stop - start));
        }

        /// <summary>
        /// Returns the URL linked in a HTML anchor/link element
        /// </summary>
        /// <param name="text">Full HTML Anchor/link element</param>
        /// <returns>The contained URL</returns>
        public string GetLinkedURL(string text)
        {
            return GetLinkedURL(text, _tagDelimiter);
        }

        /// <summary>
        /// Returns the URL linked in a HTML anchor/link element
        /// </summary>
        /// <param name="text">Full HTML Anchor/link element</param>
        /// <returns>The contained URL</returns>
        public string GetLinkedURL(string text, string tagDelimiter)
        {
            string matchtext = text.ToLower();

            int start = matchtext.IndexOf("href=" + tagDelimiter) + 6;
            int stop = matchtext.IndexOf(tagDelimiter, start);

            return text.Substring(start, (stop - start));
        }

        /// <summary>
        /// Finds the first table from the current position and creates
        /// a 2-dimensional string-array containing row and column-data
        /// </summary>
        /// <returns></returns>
        public string[,] GetTableData()
        {
            string tempData = _data.Substring(_position);
            int endPos = tempData.ToLower().IndexOf("</table>");
            if (endPos != -1)
            {
                tempData = tempData.Substring(0, endPos);
            }

            string tempData2 = tempData.ToLower();

            int pos, numRows, numColumns;
            pos = 0;
            numColumns = 0;

            // determine table geometry up front
            string[] rows = tempData2.Replace("<tr", "Å").Split('Å');

            var rowsFiltered = new List<string>();
            foreach (var row in rows)
            {
                if (!string.IsNullOrEmpty(row) && row.Contains("</td>"))
                {
                    rowsFiltered.Add(row);
                }
            }
            rows = rowsFiltered.ToArray();
            numRows = rows.Length;

            foreach (var row in rows)
            {
                string[] columns = row.Replace("<td", "Å").Split('Å');
                var rowColumns = columns.Length - 1; // account for splitting including one more

                if (rowColumns > numColumns)
                {
                    numColumns = rowColumns;
                }
            }

            // generate table from HTML
            string[,] tableData = new string[numRows,numColumns];

            for (int row = 0; row < numRows; row++)
            {
                //tableData[row] = new string[numColumns];
                pos = tempData2.IndexOf("<tr", pos) + 1;
                pos = tempData2.IndexOf(">", pos) + 1;

                for (int column = 0; column < numColumns; column++)
                {
                    pos = tempData2.IndexOf("<td", pos) + 1;
                    pos = tempData2.IndexOf(">", pos) + 1;

                    endPos = tempData2.IndexOf("</td>", pos);
                    var cellData = tempData.Substring(pos, (endPos - pos));
                    tableData[row,column] = cellData;
                    pos = endPos;
                }
            }

            return tableData;
        }
    }
}