﻿using System.Linq;
using System.Threading;
using Sayumin.Common;
using Sayumin.Common.DAC;
using System.Collections.Generic;
using System;
using System.Net;
using System.IO;
using System.Data.SqlClient;
using System.Configuration;

namespace Sayumin.QueueDownloader
{
    public partial class QueueDownloaderService : SayuminBaseService
    {
        protected SayuminDataContext db;

        public QueueDownloaderService()
            : base()
        {
            ServiceName = "Sayumin.QueueDownloader";
        }

        protected override void DoWork()
        {
            db = new SayuminDataContext();


            int scansPerMinute = 1;
            int totalSleepLength = (int)(60000 / scansPerMinute); // milliseconds

            int sleepStateLength = 1000; // milliseconds
            int sleepsPerScan = totalSleepLength / sleepStateLength;

            while (false == _exitSignaled)
            {
                ResetStatus();
                ProcessDownloads();

                for (int i = 0; i < sleepsPerScan; i++)
                {
                    Thread.Sleep(sleepStateLength);

                    if (true == _exitSignaled)
                    {
                        break;
                    }
                }
            }
        }

        protected void ResetStatus()
        {
            var stalled = from fq in db.FileQueues
                          where fq.StatusNumber == (int)QueueStatus.Downloading
                          select fq;

            foreach (FileQueue fq in stalled)
            {
                fq.Status = QueueStatus.QueuedForDownload;
            }

            db.SubmitChanges();
        }

        protected void ProcessDownloads()
        {
            var queue = (from fq in db.FileQueues
                         where fq.StatusNumber == (int)QueueStatus.QueuedForDownload
                         orderby fq.Attempts
                         group fq by fq.ServerID into g
                         select new
                         {
                             ServerID = g.Key,
                             QueueItems = g
                         }).ToList();

            // TODO: implement threaded model (causes DB HELL)
            //List<Thread> threads = new List<Thread>();

            foreach (var itemGroup in queue)
            {
                IEnumerable<FileQueue> items = itemGroup.QueueItems;
                ProcessQueueItems(items);

                //Thread thread = new Thread(ProcessQueueItems);
                //threads.Add(thread);
                //thread.Start(items);
            }

            //bool waitState = true;
            //while (waitState) 
            //{
            //    bool isCompleted = true;
            //    foreach (Thread thread in threads)
            //    {
            //        isCompleted = isCompleted && thread.ThreadState == ThreadState.Stopped;
            //    }
            //    waitState = !isCompleted;

            //    Thread.Sleep(1000);
            //}
        }

        protected void ProcessQueueItems(object QueueItems)
        {
            IEnumerable<FileQueue> items = (IEnumerable<FileQueue>)QueueItems;
            ProcessQueueItems(items);
        }

        protected void ProcessQueueItems(IEnumerable<FileQueue> items)
        {
            foreach (FileQueue item in items)
            {
                lock (db)
                {
                    item.Status = QueueStatus.Downloading;
                    db.SubmitChanges();
                }

                try
                {
                    ProcessQueueItem(item);
                }
                catch (Exception ex)
                {
                    string errorText = string.Format("Error processing download for radioshow {0} ({1:yyyy-MM-dd}).", item.RadioShow.DisplayName, item.EpisodeDate);
                    _logModule.WriteError(errorText, ex);

                    lock (db)
                    {
                        item.Status = QueueStatus.QueuedForDownload;
                        item.Attempts++;
                        db.SubmitChanges();
                    }
                }
            }
        }

        protected void ProcessQueueItem(FileQueue item)
        {
            string dateString = item.EpisodeDate.ToString("yyyy.MM.dd");
            string localFileName = string.Format(item.RadioShow.FileNameBase, dateString);
            string folder = item.RadioShow.Folder;

            string messages = "";
            using (MediaServerHandler msh = new MediaServerHandler())
            {
                msh.DownloadData(item.URL, folder, localFileName, item.RadioShow.RetainEpisodes);
                messages = msh.Messages;
            }

            string server = ConfigurationSettings.AppSettings["serverUrl"];
            string publicURL = string.Format("{0}/{1}/{2}", server, folder, localFileName);

            long size = 0;
            try
            {
                size = this.GetSize(publicURL);
            }
            catch (Exception ex)
            {
                string errorText = string.Format("Failed to verify download and size for show {0} ({1:yyyy-MM-dd}). SSH Log: {2}", item.RadioShow.DisplayName, item.EpisodeDate, messages);
                _logModule.WriteError(errorText, ex);

                lock (db)
                {
                    item.Status = QueueStatus.QueuedForDownload;
                    item.Attempts++;
                    db.SubmitChanges();
                }

                return;
            }

            RegisterEpisode(item, publicURL, size);
            item.Status = QueueStatus.Completed;
            db.SubmitChanges();
        }

        protected long GetSize(string url)
        {
            WebClient wc = new WebClient();

            Stream stream = wc.OpenRead(url);
            WebHeaderCollection headers = wc.ResponseHeaders;

            long size = 0;
            for (int i = 0; i < headers.Count; i++)
            {
                if (headers.GetKey(i).ToLower() == "content-length")
                {
                    size = Convert.ToInt64(headers.Get(i));
                    break;
                }
            }
            stream.Close();

            return size;
        }

        protected void RegisterEpisode(FileQueue item, string publicURL, long size)
        {
            // prevents polluting the main connection with incorrect/duplicate entries.
            SayuminDataContext db = new SayuminDataContext();

            // register old deleted episodes inactive
            var oldEpisodes = (from e in db.Episodes
                               where e.RadioShow == item.RadioShow
                               && e.IsAvailable == true
                               orderby e.Date descending
                               select e).Skip<Episode>(item.RadioShow.RetainEpisodes - 1);

            foreach (Episode e in oldEpisodes)
            {
                e.IsAvailable = false;
            }

            // register episode
            Episode episode = null;


            episode = new Episode()
            {
                 Date = item.EpisodeDate,
                 FetchDate = DateTime.Now,
                 IsAvailable = true,
                 PublicURL = publicURL,
                 RadioShowID = item.RadioShowID,
                 ScheduleID = item.ScheduleID,
                 Size = size
            };

            db.Episodes.InsertOnSubmit(episode);

            // submit
            bool isDuplicate = db.SubmitChangesAndIgnoreDuplicates();

            if (isDuplicate)
            {
                // refresh existing entry with current-data
                SayuminDataContext newDb = new SayuminDataContext();
                Episode oldEntry = (from e in newDb.Episodes
                                    where e.RadioShowID == episode.RadioShowID
                                    && e.Date == episode.Date
                                    select e).Single();

                oldEntry.FetchDate = episode.FetchDate;
                oldEntry.Size = episode.Size;
                oldEntry.FetchDate = DateTime.Now;

                newDb.SubmitChanges();
            }

            string logMessage = string.Format("Radioshow {0} ({1:yyyy-MM-dd}) downloaded and registered successfully.", item.RadioShow.DisplayName, item.EpisodeDate);
            _logModule.Write(logMessage);
        }
    }
}