﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using Sayumin.Common;
using Sayumin.Common.DAC;

namespace Sayumin.QueueProcessor
{
    public partial class QueueProcessorService : SayuminBaseService
    {
        protected SayuminDataContext db;

        protected Dictionary<int, IFileSource> _sources;
        protected Dictionary<int, FileEntry[]> _sourceData;

        protected const int CACHE_DURATION = 10; // minutes

        public QueueProcessorService() : base()
        {
            ServiceName = "Sayumin.QueueProcessor";

            db = new SayuminDataContext();
            _sources = new Dictionary<int, IFileSource>();
            _sourceData = new Dictionary<int, FileEntry[]>();
        }

        protected override void DoWork()
        {
            int scansPerMinute = 1;
            int totalSleepLength = (int)(60000 / scansPerMinute); // milliseconds

            int sleepStateLength = 1000; // milliseconds
            int sleepsPerScan = totalSleepLength / sleepStateLength;

            while (false == _exitSignaled)
            {
                for (int i = 0; i < sleepsPerScan; i++)
                {
                    Thread.Sleep(sleepStateLength);

                    if (true == _exitSignaled)
                    {
                        break;
                    }
                }

                // due to uknown main-loop null-errors, do sleep before scan not to throttle sources
                ProcessSchedule();
            }
        }

        protected void ProcessSchedule()
        {
            var queue = from fq in db.FileQueues
                        where fq.StatusNumber == (int)QueueStatus.Active
                        select fq;

            foreach (FileQueue item in queue)
            {
                ProcessQueueItem(item);
            }
        }

        protected void ProcessQueueItem(FileQueue Item)
        {
            FileEntry match = null;
            Source matchSource = null;

            try
            {
                // tag busy
                Item.Status = QueueStatus.Processing;
                db.SubmitChanges();

                // process data
                var sourceConfigurations = from rss in Item.RadioShow.RadioshowSources
                                           orderby rss.Priority
                                           select rss;

                foreach (RadioshowSource rss in sourceConfigurations)
                {
                    FileEntry[] entries = GetSourceData(rss);
                    IRadioScanner scanner = GetRadioScanner(rss);
                    match = scanner.FindMatch(entries, Item.EpisodeDate);

                    if (match != null)
                    {
                        matchSource = rss.Source;
                        break;
                    }
                }

                if (match != null)
                {
                    Item.ServerID = matchSource.ServerID;
                    Item.URL = match.Url;
                    Item.UploadDate = match.UploadDate;
                    Item.Status = QueueStatus.QueuedForDownload;
                }
                else
                {
                    Item.Attempts++;
                    Item.Status = QueueStatus.Active;
                }
                db.SubmitChanges();
            }
            catch (Exception ex)
            {
                // ensure that errors does not kill queue-items from being processed later.
                Item.Status = QueueStatus.Active;
                db.SubmitChanges();

                throw ex;
            }

            // defer logging until comitted
            if (match != null)
            {
                string logMessage = string.Format("Radioshow {0} ({1:yyyy-MM-dd}) located and queued for download.", Item.RadioShow.DisplayName, Item.EpisodeDate);
                _logModule.Write(logMessage);
            }
        }

        protected FileEntry[] GetSourceData(RadioshowSource rss)
        {
            Source sourceSettings = rss.Source;
            if (false == _sources.Keys.Contains(sourceSettings.ID))
            {
                IFileSource newSource = sourceSettings.GetInstance();
                
                IConfigurable configurable = newSource as IConfigurable;
                if (configurable != null)
                {
                    configurable.InitSettings(sourceSettings.Configuration);
                }

                _sources[sourceSettings.ID] = newSource;
            }

            IFileSource source = _sources[sourceSettings.ID];

            DateTime now = DateTime.Now;
            DateTime cacheExpires = source.LastUpdated.AddMinutes(CACHE_DURATION);
            bool cacheExpired = now > cacheExpires;

            if (cacheExpired || false == _sourceData.Keys.Contains(sourceSettings.ID))
            {
                FileEntry[] newEntries = source.GetEntries();
                _sourceData[sourceSettings.ID] = newEntries;
            }

            FileEntry[] entries = _sourceData[sourceSettings.ID];
            return entries;
        }

        protected IRadioScanner GetRadioScanner(RadioshowSource rss)
        {
            IRadioScanner scanner = rss.Scanner.GetInstance();

            IConfigurable configurable = scanner as IConfigurable;
            if (configurable != null)
            {
                configurable.InitSettings(rss.Configuration);
            }

            return scanner;
        }
    }
}
