﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Sayumin.Common.DAC;

namespace Sayumin.Tests.SourceConfig
{
    public class ScheduleController : IScheduleController
    {
        protected IScheduleView _view;
        protected RadioShow _show;
        protected SourceConfiguration _configurator;
        protected SourceFileEntry[] _matches;
        protected SayuminDataContext _db;

        public ScheduleController()
        {
            _configurator = new SourceConfiguration();
            _db = new SayuminDataContext();
        }

        #region IScheduleController Members

        public IScheduleView View
        {
            get
            {
                return _view;
            }
            set
            {
                _view = value;
            }
        }

        public RadioShow RadioShow
        {
            get
            {
                return _show;
            }
            set
            {
                _show = value;
                UpdateView();
            }
        }

        public void DateChosen()
        {
            DayOfWeek dayOfWeek = _view.DatePicked.DayOfWeek;

            Schedule matchedSchedule = null;
            foreach (Schedule s in _show.Schedules)
            {
                if (s.DayOfWeekAired == dayOfWeek)
                {
                    matchedSchedule = s;
                    break;
                }
            }

            if (matchedSchedule == null)
            {
                Schedule newSchedule = new Schedule()
                {
                    DayOfWeek = (int)dayOfWeek,
                    IsActive = _view.IsActive,
                    RadioShow = _show,
                    StartDay = (int)dayOfWeek,
                    StartTime = _view.TimePicked
                };

                _show.Schedules.Add(newSchedule);
            }
            else
            {
                _db.Schedules.DeleteOnSubmit(matchedSchedule);
                _show.Schedules.Remove(matchedSchedule);
            }

            UpdateView();
        }

        public void DayChosen()
        {
            bool isActive = _view.IsActive;
            DateTime startTime = _view.TimePicked;

            DayOfWeek[] activeDays = _view.ActiveDays;

            _show.Schedules.Clear();

            foreach (DayOfWeek day in activeDays)
            {
                Schedule newSchedule = new Schedule()
                {
                    DayOfWeek = (int)day,
                    IsActive = _view.IsActive,
                    RadioShow = _show,
                    StartDay = (int)day,
                    StartTime = _view.TimePicked
                };

                _show.Schedules.Add(newSchedule);
            }

            UpdateView();
        }

        public void EnabledChosen()
        {
            bool isActive = _view.IsActive;

            foreach (Schedule s in _show.Schedules)
            {
                s.IsActive = isActive;
            }

            UpdateView();
        }

        public void StartTimeChanged()
        {
            foreach (Schedule s in _show.Schedules)
            {
                s.StartTime = _view.TimePicked;
            }
        }

        #endregion

        protected void UpdateView()
        {
            bool isActive = _show.Schedules.Count == 0 ? true : _show.Schedules[0].IsActive;
            _view.IsActive = isActive;

            DayOfWeek[] activeDays = (from s in _show.Schedules
                                      select s.DayOfWeekAired).ToArray();
            _view.ActiveDays = activeDays;

            DateTime startTime = _show.Schedules.Count == 0 ? DateTime.Now : _show.Schedules[0].StartTime;
            _view.TimePicked = startTime;

            List<DateTime> dates = GetRecentDates();

            List<string> datesFormatted = new List<string>();
            foreach (DateTime date in dates)
            {
                datesFormatted.Add(date.ToString("yyyy-MM-dd"));
            }
            _view.RecentDates = datesFormatted.ToArray();

            _view.Filters = null;
            _view.Filters = _show.RadioshowSources;
        }

        protected List<DateTime> GetRecentDates()
        {
            List<DateTime> dates;
            dates = new List<DateTime>();
            foreach (Schedule s in _show.Schedules)
            {
                DateTime date = GetDate(s.DayOfWeekAired);
                for (int i = 0; i < 4; i++)
                {
                    dates.Add(date);
                    date = date.AddDays(-7);
                }
            }
            dates.Sort();
            dates.Reverse();
            return dates;
        }

        protected DateTime GetDate(DayOfWeek Day)
        {
            DateTime date = DateTime.Now;

            while (date.DayOfWeek != Day)
            {
                date = date.AddDays(-1);
            }

            return date;
        }

        public void TestFilters()
        {
            string description = _view.Description;
            string extension = _view.Extension;
            string orig = _view.Orig;

            DateTime[] dates = GetRecentDates().ToArray();

            _matches = _configurator.GetMatchedFiles(description, extension, orig, dates);

            _view.TestResults = _matches;
        }

        public void AddFilters()
        {
            TimeSpan minTime = TimeSpan.MaxValue;

            foreach (SourceFileEntry entry in _matches)
            {
                TimeSpan uploaded = entry.UploadDate.TimeOfDay;

                if (uploaded < minTime)
                {
                    minTime = uploaded;
                }
            }

            string description = _view.Description;
            string extension = _view.Extension;
            string orig = _view.Orig;

            string configString = _configurator.GetConfigurationString(description, extension, orig);

            var sources = (from s in _matches
                           select s.Source).Distinct();

            foreach (Source source in sources)
            {
                RadioshowSource rss = new RadioshowSource()
                {
                    Configuration = configString,
                    Priority = 5,
                    RadioShow = _show,
                    ScannerID = 3,
                    SourceID = source.ID
                };

                _show.RadioshowSources.Add(rss);
            }

            UpdateView();
        }

        public void New()
        {
            _show = new RadioShow();
            // reset all changes to old datacontext and shows
            _db = new SayuminDataContext();
            UpdateView();
        }

        public void Open()
        {
            RadioShow[] shows = (from rs in _db.RadioShows
                                 orderby rs.DisplayName
                                 select rs).ToArray();
            OpenRadioshowForm orf = new OpenRadioshowForm();
            orf.PopulateRadioShows(shows);
            orf.ShowDialog();

            if (orf.SelectedShow != null)
            {
                // reset all changes to old datacontext and shows
                _db = new SayuminDataContext();

                _show = (from rs in _db.RadioShows
                         where rs.ID == orf.SelectedShow.ID
                         select rs).Single();
                UpdateView();
            }
        }

        public void Save()
        {
            if (_show.ID != 0)
            {
                _db.SubmitChanges();
                _db = new SayuminDataContext();
            }
            else
            {
                SaveRadioshowForm srf = new SaveRadioshowForm();
                srf.ShowDialog();

                if (srf.SaveConfirmed)
                {
                    _show.DisplayName = srf.DisplayName;
                    _show.Folder = srf.FolderName;
                    _show.FileNameBase = srf.FilenameBase;
                    _show.RetainEpisodes = srf.RetainEpisodes;
                    _show.Text = "<p>No description added yet.</p>";

                    _db.RadioShows.InsertOnSubmit(_show);
                    _db.SubmitChanges();
                }
            }
        }
    }
}
