﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Sayumin.Common.DAC;
using Sayumin.Common;

namespace Sayumin.Tests.SourceConfig
{
    #region SourceFileEntry class

    public class SourceFileEntry : FileEntry
    {
        protected Source _source;

        public Source Source
        {
            get { return _source; }
            set { _source = value; }
        }

        // for UI databinding
        public string SourceName
        {
            get { return _source.DisplayName; }
        }
        
        public SourceFileEntry(Source Source, FileEntry Entry)
        {
            Columns = Entry.Columns;
            Description = Entry.Description;
            Extension = Entry.Extension;
            FileSize = Entry.FileSize;
            UploadDate = Entry.UploadDate;
            Url = Entry.Url;
            _source = Source;
        }

    }

    #endregion SourceFileEntry class

    public class SourceConfiguration
    {
        protected SayuminDataContext _db;
        protected Dictionary<Source, FileEntry[]> _sourceData;
        protected List<SourceFileEntry> _sourceFileEntries;

        public SourceConfiguration()
        {
            _sourceData = new Dictionary<Source, FileEntry[]>();
            _sourceFileEntries = new List<SourceFileEntry>();
            _db = new SayuminDataContext();
        }

        protected void LoadSourceData()
        {
            int failedSources = 0;

            if (_sourceData.Keys.Count != 0)
            {
                return;
            }

            ProgressDialog pd = new ProgressDialog();
            pd.Text = "Getting source-data";
            pd.Show();

            var sources = (from s in _db.Sources
                          select s).ToList();

            int sourceNum = 0;

            foreach (Source source in sources)
            {
                int progress = Convert.ToInt32(100 * sourceNum / sources.Count);
                pd.Progress = progress;

                IFileSource instance = source.GetInstance();
                IConfigurable configurableModule = instance as IConfigurable;
                if (configurableModule != null)
                {
                    configurableModule.InitSettings(source.Configuration);
                }

                try
                {
                    FileEntry[] fileEntries = instance.GetEntries();
                    _sourceData[source] = fileEntries;
                }
                catch (NullReferenceException)
                {
                    // dont pay attention to failed sources
                    failedSources++;
                }

                sourceNum++;
            }
            // we just need a breakpoint;
            failedSources = failedSources + 1 - 1;

            pd.Close();
        }

        public SourceFileEntry[] GetFileEntries()
        {
            if (_sourceFileEntries.Count != 0)
            {
                return _sourceFileEntries.ToArray();
            }

            LoadSourceData();

            foreach (Source source in _sourceData.Keys)
            {
                _sourceFileEntries.AddRange(
                    (from fe in _sourceData[source]
                    select new SourceFileEntry(source, fe)).ToList()
                    );
            }

            return _sourceFileEntries.ToArray();
        }

        public SourceFileEntry[] GetMatchedFiles(string description, string extension, string orig, DateTime[] dates)
        {
            // assumes extendedscanner

            string configString = GetConfigurationString(description, extension, orig);
            IRadioScanner scanner = GetScanner(configString);

            LoadSourceData();

            List<SourceFileEntry> entries = new List<SourceFileEntry>();
            foreach (Source source in _sourceData.Keys)
            {
                FileEntry[] fileEntries = _sourceData[source];

                foreach (DateTime date in dates)
                {
                    FileEntry matchedEntry = scanner.FindMatch(fileEntries, date);

                    if (matchedEntry != null)
                    {
                        SourceFileEntry match = new SourceFileEntry(source, matchedEntry);
                        entries.Add(match);
                    }
                }
            }

            return entries.ToArray();
        }

        protected IRadioScanner GetScanner(string configString)
        {
            Scanner scanner = (from s in _db.Scanners
                               where s.DisplayName == "ExtendedScanner"
                               select s).Single();

            IRadioScanner instance = scanner.GetInstance();
            IConfigurable configurableModule = instance as IConfigurable;
            if (configurableModule != null)
            {
                configurableModule.InitSettings(configString);
            }

            return instance;
        }

        public string GetConfigurationString(string description, string extension, string orig)
        {
            string configString = string.Empty;

            if (false == string.IsNullOrEmpty(description))
            {
                configString += ";desc=" + description;
            }

            if (false == string.IsNullOrEmpty(extension))
            {
                configString += ";ext=" + extension;
            }

            //if (false == string.IsNullOrEmpty(orig))
            //{
            //    configString += ";matchKey=orig;matchValue" + orig;
            //}

            if (configString.Length != 0)
            {
                configString = configString.Substring(1);
            }

            return configString;
        }
    }
}
